#include "pcr3bp.h"
#include "PMap.h"
// #####################################################################
/**
 * This function checks the cone condtion of a given set. 
 * @param[in] N - h set
*/
void checkConeCondition(PMap& pm, TripleSet &H, string message)
{
	writeReport("Cone condition on " + message + ": ");
  IVector u = H.Icenter + interval(-1,1)*H.Iunstable + interval(-1,1)*H.Istable;
  IMatrix DP = pm.derivative(u,2,H.sign);

  IMatrix Q(2,2); Q[0][0] = .0001; Q[1][1]=-1;
  IMatrix C(2,2);
  C.column(0) = H.Iunstable;
  C.column(1) = H.Istable;

  std::ofstream out;
  std::string filename = message+".dat"; 
  out.open(filename);
  out.precision(17);
  out << "quadratic form Q = " << Q << endl;
  out << "coordinates of the set C = " << C << endl;  
  out << "DP in main coordinates=" << DP << endl;
  DP = capd::matrixAlgorithms::krawczykInverse(C)*DP*C;
  out << "DP in coordinates of the set, i.e. V = C^{-1}*DP*C = " << DP << endl;  
  out << "The matrix V^T*Q*V-Q should be positive definite:\n";
  IMatrix V = Transpose(DP)*Q*DP-Q;
  out << "V = " << V << endl;
  out << "From Sylvester criterion we should have V(0,0)>0 and det(V)>0. We have\n";
  interval z;
  intersection(V[0][1],V[1][0],z);
  out << "V(0,0)=" << V[0][0] << endl;
  out << "det(V) = " << V[0][0]*V[1][1] - sqr(z) << endl;
  out.close();
  bool result = V[0][0] > 0 and (V[0][0]*V[1][1] - sqr(z) > 0);
  cout << (result ? "done" : "failed") << " (See file " << filename  << " for details)" << endl;
}


// #####################################################################
/**
 * This function checks if the condition holds true for a Poincare map of the set covered by a grid
 * @param[in] grid - grid of a domain
 * @param[in] Q,y - coordinate system in the image
 * @parampin] iterations - number of iterations of the Poincare map
 * @param[in] sign - sign of the section 
 * @param[in] condition to be satisfied
 * @return true if condition satisfied
 */ 
template<class Condition>
bool checkCondition(PMap& pm, GridSet& grid, capd4::IMatrix Q, capd4::IVector y, int iterations, double sign, Condition condition, std::string message=""){
  GridSet::iterator b = grid.begin();
  std::list< std::pair<capd4::IVector,capd4::IVector> > unchecked;
  while(b!=grid.end()){
    unchecked.push_back(std::make_pair(*b,grid.r));
    ++b;
  }
  
  const int loop_limit = 400;
  
  while(!unchecked.empty() and unchecked.size()<loop_limit){
    auto elem = unchecked.front();
    unchecked.pop_front();
    capd4::IVector v = elem.first + grid.C*elem.second;
    elem.first[3] = sign*pm.embeding(v);
    elem.first[3].split(elem.second[3]);
    IVector im = pm(elem.first,grid.C,elem.second,Q,y,iterations);
    if(!condition(im)){
      if(unchecked.size()+2>=loop_limit){
        writeReport("\n"+message+" (Inclusion not satisfied) ");
        std::cout << im << std::flush;
      }
      // subdivision of element
      elem.second[3] = 0.;
      capd4::IVector m = midVector(elem.second);
      capd4::IVector x1 = intervalHull(leftVector(elem.second),m);
      capd4::IVector x2 = intervalHull(rightVector(elem.second),m);
      capd4::IVector r1(4), r2(4);
      split(x1,r1);
      split(x2,r2);
      unchecked.push_back(std::make_pair(elem.first + grid.C*x1,r1));
      unchecked.push_back(std::make_pair(elem.first + grid.C*x2,r2));
    }
  }
  return unchecked.empty();
}

// #####################################################################
/**
 * This function checks covering relation between two sets
 * @param[in] N,M - h-sets
 * @parampin] iterations - number of iterations of the Poincare map
 * @param[in] sign - sign of the section 
 * @return true if covering relation is validated
 */ 
bool checkCovering(PMap& pm, TripleSet& N, TripleSet& M, string message, int iterations){
  writeReport(message);
  bool result = true;
  GridSet grid(4);
  capd4::IMatrix Q = pm.embed(M.Iunstable,M.Istable);
  Q = capd::matrixAlgorithms::krawczykInverse(Q);
  capd4::IVector y = pm.embed(M.Icenter,M.sign);

  auto left = [](const IVector& x){return x[0]<-1.;};
  auto right = [](const IVector& x){return x[0]>1.;};
  auto across = [](const IVector& x){return capd::abs(x[1])<1. or x[0]>1. or x[0]<-1.;};

  const int initHorizontalGridSize = 5;
  const int initVerticalGridSize = 5;
  N.gridTopEdge(grid,initHorizontalGridSize,0,2);
  result = result and checkCondition(pm,grid,Q,y,iterations,N.sign,across,"top ");
  N.gridBottomEdge(grid,initHorizontalGridSize,0,2);
  result = result and checkCondition(pm,grid,Q,y,iterations,N.sign,across,"bottom ");
  N.gridRightEdge(grid,initVerticalGridSize,0,2);
  result = result and checkCondition(pm,grid,Q,y,iterations,N.sign,right,"right ");
  N.gridLeftEdge(grid,initVerticalGridSize,0,2);
  result = result and checkCondition(pm,grid,Q,y,iterations,N.sign,left,"left ");
  
  // existence of Poincare map
  // no subdivision (1x1)
  // no restrictions on the image
  auto alwaysTrue = [](const IVector&) { return true; };
  N.gridSet(grid,1,1,0,2);
  checkCondition(pm,grid,Q,y,iterations,N.sign,alwaysTrue,"exists ");
  writeReport(result ? ": done\n" : ": failed\n");
  
  return result;
}

